#include <proto.h>

#include "gtgc.h"
#include "gterror.h"

/************************************************************************\
*
* GtGC class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTGC_MASK	 (GCFunction|GCPlaneMask|GCForeground|GCBackground|\
										GCLineWidth|GCLineStyle|GCCapStyle|GCJoinStyle|\
										GCFillStyle|GCFillRule|GCTileStipXOrigin|\
										GCTileStipYOrigin|GCSubwindowMode|GCGraphicsExposures|\
										GCClipXOrigin|GCClipYOrigin|GCDashOffset|GCArcMode)

static GtGC *first=NULL;

/************************************************************************\
*
* Create and init gc.
*
\************************************************************************/
GtGC *GtMkGC(GtDisplay *d,GtTag *tgs)
{
	GtGC *obj;
	
	if(!(obj=new GtGC()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtGC object!\n");
		return(NULL);
	}
	if(!obj->Init(d,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtGC *GtMkGC(GtDisplay *d,ulong tag,...)
{
	va_list va;
	GtGC *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkGC(d,GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtGC *GtMkSharedGC(GtDisplay *d,GtTag *tgs)
{
	GtTag *t,*tt;
	ulong gmask=0;
	XGCValues gattr,gattr2;
	GtPixmap *gtile=NULL;
	GtPixmap *gstip=NULL;
	GtFont *gfont=NULL;
	GtPixmap *gcmask=NULL;
	GtDisplay *gdsp=NULL;
	GtDrawable *gdr=NULL;
	GtGC *obj,*g;

	memset(&gattr,0,sizeof(XGCValues));
	memset(&gattr2,0,sizeof(XGCValues));
	gdsp=d;
	tt=tgs;
	while((t=GtNextTag(&tt)))
	{
		switch(t->tag)
		{
			case GT_Drawable:
				gdr=(GtDrawable *)t->d.ptr;
				break;
			case GT_Function:
				gattr.function=t->d.val;
				gmask|=GCFunction;
				break;
			case GT_PlaneMask:
				gattr.plane_mask=t->d.val;
				gmask|=GCPlaneMask;
				break;
			case GT_Foreground:
				gattr.foreground=t->d.val;
				gmask|=GCForeground;
				break;
			case GT_Background:
				gattr.background=t->d.val;
				gmask|=GCBackground;
				break;
			case GT_LineWidth:
				gattr.line_width=t->d.val;
				gmask|=GCLineWidth;
				break;
			case GT_LineStyle:
				gattr.line_style=t->d.val;
				gmask|=GCLineStyle;
				break;
			case GT_CapStyle:
				gattr.cap_style=t->d.val;
				gmask|=GCCapStyle;
				break;
			case GT_JoinStyle:
				gattr.join_style=t->d.val;
				gmask|=GCJoinStyle;
				break;
			case GT_FillStyle:
				gattr.fill_style=t->d.val;
				gmask|=GCFillStyle;
				break;
			case GT_FillRule:
				gattr.fill_rule=t->d.val;
				gmask|=GCFillRule;
				break;
			case GT_Tile:
				if((gtile=(GtPixmap *)t->d.ptr))
					gattr.tile=gtile->XPix();
				else
					gattr.tile=None;
				gmask|=GCTile;
				break;
			case GT_Stipple:
				if((gstip=(GtPixmap *)t->d.ptr))
					gattr.stipple=gstip->XPix();
				else
					gattr.stipple=None;
				gmask|=GCStipple;
				break;
			case GT_TStipXOrigin:
				gattr.ts_x_origin=t->d.val;
				gmask|=GCTileStipXOrigin;
				break;
			case GT_TStipYOrigin:
				gattr.ts_y_origin=t->d.val;
				gmask|=GCTileStipYOrigin;
				break;
			case GT_Font:
				if((gfont=(GtFont *)t->d.ptr))
					gattr.font=gfont->XFnt();
				else
					gattr.font=None;
				gmask|=GCFont;
				break;
			case GT_SubWinMode:
				gattr.subwindow_mode=t->d.val;
				gmask|=GCSubwindowMode;
				break;
			case GT_GraphExpose:
				gattr.graphics_exposures=t->d.val;
				gmask|=GCGraphicsExposures;
				break;
			case GT_ClipXOrigin:
				gattr.clip_x_origin=t->d.val;
				gmask|=GCClipXOrigin;
				break;
			case GT_ClipYOrigin:
				gattr.clip_y_origin=t->d.val;
				gmask|=GCClipYOrigin;
				break;
			case GT_ClipMask:
				if((gcmask=(GtPixmap *)t->d.ptr))
					gattr.clip_mask=gcmask->XPix();
				else
					gattr.clip_mask=None;
				gmask|=GCClipMask;
				break;
			case GT_DashOffset:
				gattr.dash_offset=t->d.val;
				gmask|=GCDashOffset;
				break;
			case GT_DashList:
				gattr.dashes=t->d.val;
				gmask|=GCDashList;
				break;
			case GT_ArcMode:
				gattr.arc_mode=t->d.val;
				gmask|=GCArcMode;
				break;
		}
	}
	for(g=first; g; g=g->next)
	{
		if((gdsp==g->dsp) && (gdr==g->dr))
		{
			XGetGCValues(g->dsp->XDsp(),g->gc,gmask,&gattr2);
			if(memcmp(&gattr,&gattr2,sizeof(XGCValues))==0)
			{
				g->cnt++;
				return(g);
			}
		}
	}
	if(!(obj=new GtGC()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtGC object!\n");
		return(NULL);
	}
	obj->next=first;
	first=obj;
	obj->cnt=0;
	if(!obj->Init(d,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtGC *GtMkSharedGC(GtDisplay *d,ulong tag,...)
{
	va_list va;
	GtGC *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkSharedGC(d,GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Delete shared font.
*
\************************************************************************/
void GtGC::DeInit(void)
{
	GtGC *t;
	
	if(cnt>0)
	{
		cnt--;
		return;
	}
	if(cnt==0)
	{
		t=first;
		if(t==this)
			first=next;
		else
		{
			while(t && (t->next!=this))
				t=t->next;
			if(t)
				t->next=next;
		}
	}
	GtBase::DeInit();
}

/************************************************************************\
*
* Setup empty gc.
*
\************************************************************************/
GtGC::GtGC(void)
{
	next=NULL;
	dsp=NULL;
	dr=NULL;
	gc=0;
	tile=NULL;
	stip=NULL;
	font=NULL;
	cmask=NULL;
	cnt=-1;
}

/************************************************************************\
*
* Destroy gc.
*
\************************************************************************/
void GtGC::Destroy(void)
{
	if(gc)
		XFreeGC(dsp->XDsp(),gc);
	GtBase::Destroy();
}

/************************************************************************\
*
* Setup gc.
*
\************************************************************************/
sbool GtGC::Init(GtDisplay *d,GtTag *tgs)
{
	GtTag *t;
	XGCValues attr;
	ulong mask=0;

	dsp=d;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Drawable:
				dr=(GtDrawable *)t->d.ptr;
				break;
			case GT_Function:
				attr.function=t->d.val;
				mask|=GCFunction;
				break;
			case GT_PlaneMask:
				attr.plane_mask=t->d.val;
				mask|=GCPlaneMask;
				break;
			case GT_Foreground:
				attr.foreground=t->d.val;
				mask|=GCForeground;
				break;
			case GT_Background:
				attr.background=t->d.val;
				mask|=GCBackground;
				break;
			case GT_LineWidth:
				attr.line_width=t->d.val;
				mask|=GCLineWidth;
				break;
			case GT_LineStyle:
				attr.line_style=t->d.val;
				mask|=GCLineStyle;
				break;
			case GT_CapStyle:
				attr.cap_style=t->d.val;
				mask|=GCCapStyle;
				break;
			case GT_JoinStyle:
				attr.join_style=t->d.val;
				mask|=GCJoinStyle;
				break;
			case GT_FillStyle:
				attr.fill_style=t->d.val;
				mask|=GCFillStyle;
				break;
			case GT_FillRule:
				attr.fill_rule=t->d.val;
				mask|=GCFillRule;
				break;
			case GT_Tile:
				if((tile=(GtPixmap *)t->d.ptr))
					attr.tile=tile->XPix();
				else
					attr.tile=None;
				mask|=GCTile;
				break;
			case GT_Stipple:
				if((stip=(GtPixmap *)t->d.ptr))
					attr.stipple=stip->XPix();
				else
					attr.stipple=None;
				mask|=GCStipple;
				break;
			case GT_TStipXOrigin:
				attr.ts_x_origin=t->d.val;
				mask|=GCTileStipXOrigin;
				break;
			case GT_TStipYOrigin:
				attr.ts_y_origin=t->d.val;
				mask|=GCTileStipYOrigin;
				break;
			case GT_Font:
				if((font=(GtFont *)t->d.ptr))
					attr.font=font->XFnt();
				else
					attr.font=None;
				mask|=GCFont;
				break;
			case GT_SubWinMode:
				attr.subwindow_mode=t->d.val;
				mask|=GCSubwindowMode;
				break;
			case GT_GraphExpose:
				attr.graphics_exposures=t->d.val;
				mask|=GCGraphicsExposures;
				break;
			case GT_ClipXOrigin:
				attr.clip_x_origin=t->d.val;
				mask|=GCClipXOrigin;
				break;
			case GT_ClipYOrigin:
				attr.clip_y_origin=t->d.val;
				mask|=GCClipYOrigin;
				break;
			case GT_ClipMask:
				if((cmask=(GtPixmap *)t->d.ptr))
					attr.clip_mask=cmask->XPix();
				else
					attr.clip_mask=None;
				mask|=GCClipMask;
				break;
			case GT_DashOffset:
				attr.dash_offset=t->d.val;
				mask|=GCDashOffset;
				break;
			case GT_DashList:
				attr.dashes=t->d.val;
				mask|=GCDashList;
				break;
			case GT_ArcMode:
				attr.arc_mode=t->d.val;
				mask|=GCArcMode;
				break;
		}
	}
	if(!dsp)
	{
		GtError(this,GTEC_NODISPLAY,"No GtDisplay!\n");
		return(FALSE);
	}
	xdsp=dsp->XDsp();
	if(!dr)
		gc=XCreateGC(dsp->XDsp(),dsp->XDefaultRootWindow(),mask,&attr);
	else
		gc=XCreateGC(dsp->XDsp(),dr->XDr(),mask,&attr);
	if(!gc)
	{
		GtError(this,GTEC_NOXRESOURCE,"No X-GC!\n");
		return(FALSE);
	}
	return(TRUE);
}

/************************************************************************\
*
* Find display tag and then setup normal.
*
\************************************************************************/
sbool GtGC::Init(GtTag *tgs)
{
	GtDisplay *d;
	
	d=(GtDisplay *)GtGetTagPtr(tgs,GT_Display,NULL);
	return(this->Init(d,tgs));
}

/************************************************************************\
*
* Set gc attributes.
*
\************************************************************************/
sbool GtGC::Set(GtTag *tgs)
{
	GtTag *t;
	XGCValues attr;
	ulong mask=0;
	
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Function:
				attr.function=t->d.val;
				mask|=GCFunction;
				break;
			case GT_PlaneMask:
				attr.plane_mask=t->d.val;
				mask|=GCPlaneMask;
				break;
			case GT_Foreground:
				attr.foreground=t->d.val;
				mask|=GCForeground;
				break;
			case GT_Background:
				attr.background=t->d.val;
				mask|=GCBackground;
				break;
			case GT_LineWidth:
				attr.line_width=t->d.val;
				mask|=GCLineWidth;
				break;
			case GT_LineStyle:
				attr.line_style=t->d.val;
				mask|=GCLineStyle;
				break;
			case GT_CapStyle:
				attr.cap_style=t->d.val;
				mask|=GCCapStyle;
				break;
			case GT_JoinStyle:
				attr.join_style=t->d.val;
				mask|=GCJoinStyle;
				break;
			case GT_FillStyle:
				attr.fill_style=t->d.val;
				mask|=GCFillStyle;
				break;
			case GT_FillRule:
				attr.fill_rule=t->d.val;
				mask|=GCFillRule;
				break;
			case GT_Tile:
				if((tile=(GtPixmap *)t->d.ptr))
					attr.tile=tile->XPix();
				else
					attr.tile=None;
				mask|=GCTile;
				break;
			case GT_Stipple:
				if((stip=(GtPixmap *)t->d.ptr))
					attr.stipple=stip->XPix();
				else
					attr.stipple=None;
				mask|=GCStipple;
				break;
			case GT_TStipXOrigin:
				attr.ts_x_origin=t->d.val;
				mask|=GCTileStipXOrigin;
				break;
			case GT_TStipYOrigin:
				attr.ts_y_origin=t->d.val;
				mask|=GCTileStipYOrigin;
				break;
			case GT_Font:
				if((font=(GtFont *)t->d.ptr))
					attr.font=font->XFnt();
				else
					attr.font=None;
				mask|=GCFont;
				break;
			case GT_SubWinMode:
				attr.subwindow_mode=t->d.val;
				mask|=GCSubwindowMode;
				break;
			case GT_GraphExpose:
				attr.graphics_exposures=t->d.val;
				mask|=GCGraphicsExposures;
				break;
			case GT_ClipXOrigin:
				attr.clip_x_origin=t->d.val;
				mask|=GCClipXOrigin;
				break;
			case GT_ClipYOrigin:
				attr.clip_y_origin=t->d.val;
				mask|=GCClipYOrigin;
				break;
			case GT_ClipMask:
				if((cmask=(GtPixmap *)t->d.ptr))
					attr.clip_mask=cmask->XPix();
				else
					attr.clip_mask=None;
				mask|=GCClipMask;
			case GT_DashOffset:
				attr.dash_offset=t->d.val;
				mask|=GCDashOffset;
				break;
			case GT_DashList:
				attr.dashes=t->d.val;
				mask|=GCDashList;
				break;
			case GT_ArcMode:
				attr.arc_mode=t->d.val;
				mask|=GCArcMode;
				break;
		}
	}
	if(mask)
		XChangeGC(dsp->XDsp(),gc,mask,&attr);
	return(TRUE);
}

/************************************************************************\
*
* Get gc attributes.
*
\************************************************************************/
void GtGC::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	XGCValues attr;
	
	XGetGCValues(dsp->XDsp(),gc,GTGC_MASK,&attr);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Display:
				*ptr=(long)dsp;
				break;
			case GT_Function:
				*ptr=attr.function;
				break;
			case GT_PlaneMask:
				*ptr=attr.plane_mask;
				break;
			case GT_Foreground:
				*ptr=attr.foreground;
				break;
			case GT_Background:
				*ptr=attr.background;
				break;
			case GT_LineWidth:
				*ptr=attr.line_width;
				break;
			case GT_LineStyle:
				*ptr=attr.line_style;
				break;
			case GT_CapStyle:
				*ptr=attr.cap_style;
				break;
			case GT_JoinStyle:
				*ptr=attr.join_style;
				break;
			case GT_FillStyle:
				*ptr=attr.fill_style;
				break;
			case GT_FillRule:
				*ptr=attr.fill_rule;
				break;
			case GT_Tile:
				*ptr=(long)tile;
				break;
			case GT_Stipple:
				*ptr=(long)stip;
				break;
			case GT_TStipXOrigin:
				*ptr=attr.ts_x_origin;
				break;
			case GT_TStipYOrigin:
				*ptr=attr.ts_y_origin;
				break;
			case GT_Font:
				*ptr=(long)font;
				break;
			case GT_SubWinMode:
				*ptr=attr.subwindow_mode;
				break;
			case GT_GraphExpose:
				*ptr=attr.graphics_exposures;
				break;
			case GT_ClipXOrigin:
				*ptr=attr.clip_x_origin;
				break;
			case GT_ClipYOrigin:
				*ptr=attr.clip_y_origin;
				break;
			case GT_ClipMask:
				*ptr=(long)cmask;
				break;
			case GT_DashOffset:
				*ptr=attr.dash_offset;
				break;
			case GT_ArcMode:
				*ptr=attr.arc_mode;
				break;
			case GT_XGC:
				*ptr=(long)gc;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtGC::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtGC")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}
